"""pytest fixtures file for the dns_exporter project."""

import os
import subprocess
import time
from http.server import HTTPServer
from pathlib import Path
from threading import Thread

import dns.message
import httpx
import pytest
import yaml

from dns_exporter.config import Config, ConfigDict
from dns_exporter.entrypoint import main
from dns_exporter.exporter import DNSExporter
from dns_exporter.metrics import QTIME_LABELS


@pytest.fixture
def exporter():
    """Fixture to return a clean version of the DNSExporter class."""

    class CleanTestExporter(DNSExporter):
        """This is just here so tests can mess around with cls.modules without changing the global DNSExporter class."""

    CleanTestExporter.modules = None
    return CleanTestExporter


@pytest.fixture(scope="session")
def dns_exporter_no_main_no_config():
    """Run a basic server without main() and with no config."""
    print("Running server with no config on 127.0.0.1:45353 ...")
    serve_forever = HTTPServer(("127.0.0.1", 45353), DNSExporter).serve_forever
    thread = Thread(target=serve_forever)
    thread.daemon = True
    thread.start()
    time.sleep(1)
    if not thread.is_alive():
        pytest.fail("Unable to create test instance on 127.0.0.1:45353")
    yield
    print("Beginning teardown")
    # give pytest some time to flush logging from background threads,
    # so it doesn't occationally leak to stdout after the pytest result summary
    time.sleep(5)


@pytest.fixture(scope="session")
def dns_exporter_example_config():
    """Run a server with main() and with the example config."""
    print("Running server with example config on 127.0.0.1:25353 ...")
    thread = Thread(
        target=main,
        args=(["-c", "dns_exporter/dns_exporter_example.yml", "-p", "25353", "-d"],),
    )
    thread.daemon = True
    thread.start()
    time.sleep(1)
    if not thread.is_alive():
        pytest.fail("Unable to create test instance on 127.0.0.1:25353")
    yield
    print("Beginning teardown")
    # give pytest some time to flush logging from background threads,
    # so it doesn't occationally leak to stdout after the pytest result summary
    time.sleep(5)


@pytest.fixture(scope="session")
def dns_exporter_main_no_config_no_debug():
    """Run a server with main() and no config."""
    print("Running server with no config on 127.0.0.1:35353 ...")
    thread = Thread(
        target=main,
        args=(["-p", "35353", "--connection-cleanup-interval-seconds", "0"],),
    )
    thread.daemon = True
    thread.start()
    time.sleep(1)
    if not thread.is_alive():
        pytest.fail("Unable to create test instance on 127.0.0.1:35353")
    yield
    print("Beginning teardown")
    # give pytest some time to flush logging from background threads,
    # so it doesn't occationally leak to stdout after the pytest result summary
    time.sleep(5)


@pytest.fixture
def dns_exporter_param_config(request):
    """Run a server in a subprocess with the config from request.param."""
    print(f"Running dns_exporter with config {request.param} on 127.0.0.1:15353 ...")
    conf = Path(__file__).parents[0] / "tests" / "prometheus" / request.param
    proc = subprocess.Popen(
        args=["dns_exporter", "-c", str(conf), "-d"],
    )
    time.sleep(1)
    if proc.poll():
        # process didn't start properly, bail out
        pytest.fail(
            f"Unable to create test instance with config {request.param} on 127.0.0.1:15353",
        )
    yield
    print(f"Stopping dns_exporter with config {request.param} on 127.0.0.1:15353 ...")
    proc.terminate()
    time.sleep(5)


@pytest.fixture(scope="session")
def dns_exporter_example_config_connection_label():
    """Run a server with main() and with the example config with the connection label feature on."""
    print("Running server with example config and connection label feature on 127.0.0.1:15355 ...")
    os.environ["DNSEXP_CONNECTION_LABEL"] = "1"
    os.environ["COVERAGE_PROCESS_START"] = "1"
    # note: if running with stdout=subprocess.PIPE and debug mode the buffer can
    # get full which will hang the process with no further explanation.
    proc = subprocess.Popen(args=["dns_exporter", "-d", "-p", "15355"])
    time.sleep(2)
    if proc.poll():
        # process didn't start properly, bail out
        pytest.fail(
            "Unable to create test instance on 127.0.0.1:15355",
        )
    yield
    print("Stopping dns_exporter with connection label feature enabled on 127.0.0.1:15355 ...")
    proc.terminate()
    time.sleep(5)
    del os.environ["DNSEXP_CONNECTION_LABEL"]
    del os.environ["COVERAGE_PROCESS_START"]


@pytest.fixture
def dns_exporter_broken_yaml_configfile(tmp_path_factory):
    """Write a dns_exporter.yml file with invalid yaml."""
    confpath = tmp_path_factory.mktemp("conf") / "dns_exporter.yml"
    # write file to disk
    with Path.open(confpath, "w") as f:
        f.write("foo:\nbar")
    # return path to the config
    return confpath


@pytest.fixture
def dns_exporter_empty_yaml_configfile(tmp_path_factory):
    """Write a dns_exporter.yml file with no configs in it."""
    confpath = tmp_path_factory.mktemp("conf") / "dns_exporter.yml"
    # write file to disk
    with Path.open(confpath, "w") as f:
        f.write("---")
    # return path to the config
    return confpath


@pytest.fixture
def dns_exporter_invalid_yaml_configfile(tmp_path_factory):
    """Write a dns_exporter.yml file with configs with errors in it."""
    confpath = tmp_path_factory.mktemp("conf") / "dns_exporter.yml"
    # write file to disk
    with Path.open(confpath, "w") as f:
        f.write("---\n")
        f.write("modules:\n")
        f.write("  broken:\n")
        f.write("    notakey: 42\n")
    # return path to the config
    return confpath


@pytest.fixture
def prometheus_server(request, tmp_path_factory, tmpdir_factory):
    """Run a prometheus server with a config provided by the unit test in request.param."""
    # write the prometheus config with scrape configs from request.param
    targetpath = Path(__file__).parents[0] / "tests" / "prometheus" / request.param
    with Path.open(targetpath) as f:
        targets = f.read()
    targets = yaml.load(targets.encode("utf-8"), Loader=yaml.SafeLoader)
    confpath = tmp_path_factory.mktemp("prometheus") / "prometheus.yml"
    # scrape asap please
    promconf = {"global": {"scrape_interval": "1s"}}
    promconf.update(targets)
    with Path.open(confpath, "w") as f:
        f.write(yaml.dump(promconf))
    # create prometheus datadir
    prompath = tmpdir_factory.mktemp("prometheus")
    print("Running Prometheus server...")
    proc = subprocess.Popen(
        args=[
            "prometheus",
            "--config.file",
            confpath,
            "--storage.tsdb.path",
            prompath,
            "--web.listen-address",
            "127.0.0.1:9092",
        ],
    )
    if proc.poll():
        pytest.fail("Unable to start prometheus on 127.0.0.1:9091")
    print("Setup finished - prometheus is running!")

    # end buildup
    yield request.param
    # begin teardown

    print("Beginning teardown")
    print("Stopping prometheus server...")
    proc.terminate()
    proc.communicate()
    print("Teardown finished!")


@pytest.fixture
def mock_collect_zerodivisionerror(mocker):
    """Monkeypatch DNSCollector.get_dns_response() to raise ZeroDivisionError (something not explicitly handled)."""
    mocker.patch(
        "dns_exporter.collector.DNSCollector.get_dns_response",
        side_effect=ZeroDivisionError("mocked"),
    )


@pytest.fixture(scope="session")
def proxy_server():
    """Run a proxy server on localhost port 1080 using gera2ld.socks.server."""
    print("Running proxy server on 127.0.0.1:1080...")
    proc = subprocess.Popen(
        args=[
            "python3",
            "-m",
            "gera2ld.socks.server",
            "-b",
            "127.0.0.1:1080",
        ],
    )
    time.sleep(3)
    if proc.poll():
        pytest.fail("Unable to start proxy on 127.0.0.1:1080")
    print("Setup finished - proxy is running on 127.0.0.1:1080!")

    # end buildup
    yield
    # begin teardown

    print("Beginning teardown")
    print("Stopping proxy server...")
    proc.terminate()
    proc.communicate()
    print("Teardown finished!")


@pytest.fixture
def mock_dns_query_https_valuerror(mocker):
    """Monkeypatch dns.query.https to raise a ValueError like a non-2XX statuscode was received from server."""
    mocker.patch(
        "dns.query.https",
        side_effect=ValueError("mocked"),
    )


@pytest.fixture
def mock_dns_query_httpx_connecttimeout(mocker):
    """Monkeypatch dns.query.https to raise a httpx.ConnectTimeout."""
    mocker.patch(
        "dns.query.https",
        side_effect=httpx.ConnectTimeout("mocked"),
    )


@pytest.fixture
def mock_get_dns_response_tcp_eoferror(mocker):
    """Monkeypatch DNSCollector.get_dns_response_tcp() to raise EOFError."""
    mocker.patch(
        "dns.query.tcp",
        side_effect=EOFError("EOF"),
    )


@pytest.fixture
def config(exporter):
    """Return a dns_exporter.config.Config object."""
    prepared = exporter.prepare_config(
        ConfigDict(
            server="dns.google",
            query_name="example.com",
        )
    )
    return Config.create(name="test", **prepared)


@pytest.fixture
def query():
    """Return a QueryMessage."""
    return dns.message.QueryMessage(id=42)


@pytest.fixture
def labels():
    """Return a dict of labels."""
    labels: dict[str, str] = {}
    for key in [*QTIME_LABELS, "server", "ip", "port", "protocol", "family", "proxy", "query_name", "query_type"]:
        labels[key] = "none"
    return labels


@pytest.fixture
def mock_get_dns_response_connectionrefusederror(mocker):
    """Monkeypatch DNSCollector.get_dns_response() to raise ConnectionRefusedError."""
    mocker.patch(
        "dns_exporter.collector.DNSCollector.get_dns_response",
        side_effect=ConnectionRefusedError("mocked"),
    )


@pytest.fixture
def mock_get_dns_response_oserror(mocker):
    """Monkeypatch DNSCollector.get_dns_response() to raise OSError."""
    mocker.patch(
        "dns_exporter.collector.DNSCollector.get_dns_response",
        side_effect=OSError("mocked"),
    )


@pytest.fixture
def mock_dns_query_https_httx_writeerror(mocker):
    """Monkeypatch dns.query.https() to raise httpx.WriteError."""
    mocker.patch(
        "dns.query.https",
        side_effect=httpx.WriteError("mocked"),
    )


@pytest.fixture
def mock_socket_close_oserror(mocker):
    """Monkeypatch dns.query.https() to raise httpx.WriteError."""
    mocker.patch(
        "socket.socket.close",
        side_effect=OSError("mocked"),
    )



"""The ``dns_exporter.metrics`` module contains definition of all the metrics for dns_exporter.

All metrics exposed by ``dns_exporter`` are prefixed with ``dnsexp_`` (apart from ``up``
and the built-in Python metrics).
"""

from __future__ import annotations

import logging
import os

from prometheus_client.core import (
    Counter,
    Gauge,
    GaugeMetricFamily,
    Histogram,
    Info,
)
from prometheus_client.utils import INF

from dns_exporter.version import __version__

logger = logging.getLogger(f"dns_exporter.{__name__}")

########################################################
# scrape-specific metrics used by the DNSCollector (served under /query)

# the labels used in the qtime, ttl, and failure metrics
QTIME_LABELS = [
    "server",
    "ip",
    "port",
    "protocol",
    "family",
    "proxy",
    "query_name",
    "query_type",
    "transport",
    "opcode",
    "rcode",
    "flags",
    "answer",
    "authority",
    "additional",
    "nsid",
    "connection",
]
if not os.getenv("DNSEXP_CONNECTION_LABEL"):
    del QTIME_LABELS[-1]

# additional labels used in the per-RR TTL metrics
TTL_LABELS = [
    "rr_section",  # answer, authority or additional
    "rr_name",
    "rr_type",
    "rr_value",
]

FAILURE_REASONS = [
    "invalid_request_module",
    "invalid_request_config",
    "invalid_request_server",
    "invalid_request_proxy",
    "invalid_request_family",
    "invalid_request_ip",
    "invalid_request_port",
    "invalid_request_path",
    "invalid_request_protocol",
    "invalid_request_query_name",
    "invalid_request_query_type",
    "invalid_request_query_class",
    "certificate_error",
    "connection_error",
    "response_truncated",
    "socket_error",
    "timeout",
    "invalid_response_statuscode",
    "invalid_response_rcode",
    "invalid_response_flags",
    "invalid_response_answer_rrs",
    "invalid_response_authority_rrs",
    "invalid_response_additional_rrs",
    "other_failure",
]
"""FAILURE_REASONS is a list of the possible failure modes which might show up in the dnsexp_failure_reason metric."""


def get_dns_qtime_metric() -> GaugeMetricFamily:
    """``dnsexp_dns_query_time_seconds`` is the gauge used as the primary timing metric for DNS queries.

    Each DNS query duration is added to this gauge with the following labels to identify it:

        - ``server``
        - ``ip``
        - ``port``
        - ``protocol``
        - ``family``
        - ``query_name``
        - ``query_type``
        - ``transport``
        - ``opcode``
        - ``rcode``
        - ``flags``
        - ``answer``
        - ``authority``
        - ``additional``
        - ``nsid``
        - ``connection`` (if enabled)

    In some cases the ``nsid`` label has no value and the placeholder ``no_nsid`` is used instead.
    """
    return GaugeMetricFamily(
        name="dnsexp_dns_query_time_seconds",
        documentation="Gauge: DNS query time in seconds.",
        labels=QTIME_LABELS,
    )


def get_dns_success_metric(value: int | None = None) -> GaugeMetricFamily:
    """``dnsexp_dns_query_success`` is a Gauge set to 1 when a DNS query is successful, or 0 otherwise.

    A DNS query is considered failed in the following cases:
        - Configuration issues preventing a DNS query
        - Network or server issues preventing a response from reaching the exporter
        - Response parsing issues
        - Response validation issues

    A DNS query is considered a success if a DNS query response is received and any configured
    validation logic passes.
    """
    return GaugeMetricFamily(
        name="dnsexp_dns_query_success",
        documentation="Gauge: Was this DNS query successful or not, 1 for success or 0 for failure.",
        value=value,
    )


def get_dns_ttl_metric() -> GaugeMetricFamily:
    """``dnsexp_dns_response_rr_ttl_seconds`` is a Gauge which tracks the TTL of individual response RRs.

    This metric will often be set multiple times during a scrape, whenever a DNS query results in multiple
    RRs in the answer/authority/additional sections. For example, if a DNS query results in a response with
    2 ``ANSWER``, 0 ``AUTHORITY`` and 4 ``ADDITIONAL`` then this metric will be set 6 times (with different
    labels).

    This Gauge has the following labels, they are the same as ``dns_exporter.metrics.dnsexp_dns_query_time_seconds``
    plus a few more:

        - ``server``
        - ``ip``
        - ``port``
        - ``protocol``
        - ``family``
        - ``query_name``
        - ``query_type``
        - ``transport``
        - ``opcode``
        - ``rcode``
        - ``flags``
        - ``answer``
        - ``authority``
        - ``additional``
        - ``nsid``
        - ``rr_section`` (answer, authority or additional)
        - ``rr_name`` (the RR name)
        - ``rr_type`` (the RR type)
        - ``rr_value`` (the first X chars of the RR value)
        - ``connection`` (if enabled)
    """
    return GaugeMetricFamily(
        name="dnsexp_dns_response_rr_ttl_seconds",
        documentation="Gauge: DNS response RR TTL in seconds.",
        labels=[
            *QTIME_LABELS,
            *TTL_LABELS,
        ],
    )


########################################################
# exporter internal/persitent metrics (served under /metrics)

dnsexp_build_version = Info(
    name="dnsexp_build_version",
    documentation="Info: The version of dns_exporter",
)
"""``dnsexp_build_version`` is a persistent Info metric which contains the version of ``dns_exporter``.

The version is taken from the installed Python package if possible, and from _version.py written by ``setuptools_scm``
if the package is not installed, like when running from a Git checkout.
"""
dnsexp_build_version.info({"version": __version__})

dnsexp_http_requests_total = Counter(
    name="dnsexp_http_requests_total",
    documentation="Counter: The total number of HTTP requests received by this exporter since start. This counter is increased every time any HTTP request is received by the dns_exporter.",  # noqa: E501
    labelnames=["path"],
)
"""``dnsexp_http_requests_total`` is a persistent Counter keeping track of the total number of HTTP requests received
by the exporter since start.

This metric has a single label, ``path`` which is set to the request path, usually ``/query`` (for making DNS queries)
or ``/metrics`` (for getting the internal exporter metrics.
"""

dnsexp_http_responses_total = Counter(
    name="dnsexp_http_responses_total",
    documentation="Counter: The total number of HTTP responses sent by this exporter since start. This counter is increased every time an HTTP response is sent from the dns_exporter.",  # noqa: E501
    labelnames=["path", "response_code"],
)
"""``dnsexp_http_responses_total`` is a Counter keeping track of the total number of HTTP responses sent by
the exporter since start.

This metric has two labels:
    - ``path`` is set to the request path, usually ``/query`` (for making DNS queries) or
      ``/metrics`` (for getting the internal exporter metrics).
    - ``response_code`` is set to the HTTP response code, usually 200.

During normal operation the exporter will never return non-200 response codes.
"""

dnsexp_dns_queries_total = Counter(
    name="dnsexp_dns_queries_total",
    documentation="Counter: The total number of DNS queries sent by this exporter since start. This counter is increased every time the dns_exporter sends out a DNS query.",  # noqa: E501
)
"""``dnsexp_dns_queries_total`` is the Counter keeping track of how many DNS queries this exporter sends out.

This metric has no labels.
"""


dnsexp_dns_responsetime_seconds = Histogram(
    name="dnsexp_dns_responsetime_seconds",
    documentation="Histogram: DNS query response timing. This histogram is updated every time the dns_exporter receives a query response.",  # noqa: E501
    labelnames=QTIME_LABELS,
    buckets=(
        0.005,
        0.01,
        0.025,
        0.05,
        0.075,
        0.1,
        0.25,
        0.5,
        0.75,
        1.0,
        2.5,
        5.0,
        7.5,
        10.0,
        INF,
    ),
)
"""Return the dnsexp_dns_responsetime_seconds metric.

``dnsexp_dns_responsetime_seconds`` is the Histogram keeping track of how many DNS responses
this exporter received since start and how long the query took.

Each DNS query duration is observed in this histogram with the following labels to identify it:

    - ``server``
    - ``ip``
    - ``port``
    - ``protocol``
    - ``family``
    - ``query_name``
    - ``query_type``
    - ``transport``
    - ``opcode``
    - ``rcode``
    - ``flags``
    - ``answer``
    - ``authority``
    - ``additional``
    - ``nsid``

In some cases the ``nsid`` label has no value and the placeholder ``no_nsid`` is used instead.
"""


dnsexp_scrape_failures_total = Counter(
    name="dnsexp_scrape_failures_total",
    documentation="Counter: The total number of scrapes failed by failure reason, protocol, server, and proxy (where applicable). This counter is increased every time the dns_exporter receives a scrape request which fails for some reason, including response validation logic.",  # noqa: E501
    labelnames=["reason", *QTIME_LABELS],
)
"""``dnsexp_scrape_failures_total`` is the Counter keeping track of how many scrape requests failed for some reason.

This Counter has the following labels, they are the same as ``dns_exporter.metrics.dnsexp_dns_query_time_seconds``
plus the ``reason`` label which has the failure reason.

        - ``reason``
        - ``server``
        - ``ip``
        - ``port``
        - ``protocol``
        - ``family``
        - ``query_name``
        - ``query_type``
        - ``transport``
        - ``opcode``
        - ``rcode``
        - ``flags``
        - ``answer``
        - ``authority``
        - ``additional``
        - ``nsid``

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""

# socket cache stats

# labels used for socket cache metrics
SOCKET_LABELS = ["protocol", "server", "ip", "verify", "proxy"]

dnsexp_sockets_total = Gauge(
    name="dnsexp_sockets_total",
    documentation=(
        "Gauge: The number of sockets in the internal SocketCache by protocol, server, ip, verify, and proxy."
    ),
    labelnames=SOCKET_LABELS,
)
"""``dnsexp_sockets_total`` is a Gauge keeping track of the number of sockets in
the internal socket cache.

This metric has five labels:
        - ``protocol``
        - ``server``
        - ``ip``
        - ``verify`` TLS verify setting, ``True``/``False``/CA Path/``none``
        - ``proxy`` Proxy URL or ``none``

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""

dnsexp_socket_age_seconds = Gauge(
    name="dnsexp_socket_age_seconds",
    documentation=(
        "Gauge: The age in seconds of each reusable socket in the internal SocketCache by protocol, "
        "protocol, server, ip, verify, and proxy."
        "family, transport (TCP/UDP), destination ip, destination port, verify, proxy, and index."
    ),
    labelnames=[*SOCKET_LABELS, "index"],
)
"""``dnsexp_socket_age_seconds`` is a Gauge keeping track of the age of each of
the reusable sockets in the internal socket cache in the exporter.

This metric has six labels:
        - ``protocol``
        - ``server``
        - ``ip``
        - ``verify`` TLS verify setting, ``True``/``False``/CA Path/``none``
        - ``proxy`` Proxy URL or ``none``
        - ``index`` Socket index

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""

dnsexp_socket_idle_seconds = Gauge(
    name="dnsexp_socket_idle_seconds",
    documentation=(
        "Gauge: The time in seconds since the socket was last used, by protocol, family, "
        "protocol, server, ip, verify, and proxy."
        "transport (TCP/UDP), destination ip, destination port, verify, proxy, and index."
    ),
    labelnames=[*SOCKET_LABELS, "index"],
)
"""``dnsexp_socket_idle_seconds`` is a Gauge keeping track of idle time of each of
the reusable sockets in the internal socket cache in the exporter.

This metric has six labels:
        - ``protocol``
        - ``server``
        - ``ip``
        - ``verify`` TLS verify setting, ``True``/``False``/CA Path/``none``
        - ``proxy`` Proxy URL or ``none``
        - ``index`` Socket index

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""

dnsexp_socket_transmit_bytes_total = Gauge(
    name="dnsexp_socket_transmit_bytes_total",
    documentation=(
        "Gauge: The number of bytes transmitted to DNS servers by each socket. "
        "Only the raw query size is counted, excluding TCP/UDP/TLS/IP overhead. "
        "Labels for this metric are protocol, server, ip, verify, proxy, and index."
    ),
    labelnames=[*SOCKET_LABELS, "index"],
)
"""``dnsexp_socket_transmit_bytes_total`` is a Gauge keeping track of the number of bytes
transmitted to servers by each socket. Only the raw query size is counted, excluding
TCP/UDP/TLS/IP overhead.

This metric has six labels:
        - ``protocol``
        - ``server``
        - ``ip``
        - ``verify`` TLS verify setting, ``True``/``False``/CA Path/``none``
        - ``proxy`` Proxy URL or ``none``
        - ``index`` Socket index

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""

dnsexp_socket_receive_bytes_total = Gauge(
    name="dnsexp_socket_receive_bytes_total",
    documentation=(
        "Gauge: The number of bytes received from DNS servers by each socket. "
        "Only the raw response size is counted, excluding any TCP/UDP/TLS/IP overhead. "
        "Labels for this metric are protocol, server, ip, verify, proxy, and index."
    ),
    labelnames=[*SOCKET_LABELS, "index"],
)
"""``dnsexp_socket_receive_bytes_total`` is a Gauge keeping track of the number of bytes
received by each socket. Only the raw response size is counted, excluding
TCP/UDP/TLS/IP overhead.

This metric has six labels:
        - ``protocol``
        - ``server``
        - ``ip``
        - ``verify`` TLS verify setting, ``True``/``False``/CA Path
        - ``proxy`` Proxy URL or ``none``
        - ``index`` Socket index

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""

dnsexp_socket_uses_total = Gauge(
    name="dnsexp_socket_uses_total",
    documentation=(
        "Gauge: The number of times each socket has been used. "
        "Labels for this metric are protocol, server, ip, verify, proxy, and index."
    ),
    labelnames=[*SOCKET_LABELS, "index"],
)
"""``dnsexp_socket_uses_total`` is a Gauge keeping track of the number of times
each socket has been used.

This metric has six labels:
        - ``protocol``
        - ``server``
        - ``ip``
        - ``verify`` TLS verify setting, ``True``/``False``/CA Path
        - ``proxy`` Proxy URL
        - ``index`` Socket index

The placeholder ``none`` is used for cases where there is no suitable value for the label.
"""



"""Unit tests for DNSCollector and other collector.py code."""

import logging

import pytest

from dns_exporter.collector import DNSCollector
from dns_exporter.exceptions import ConfigTypeError, LabelsTypeError, QueryTypeError


def test_invalid_failure_reason(config, query):
    """Call DNSCollector.yield_failure_reason_metric() with an unknown reason."""
    c = DNSCollector(config, query, {})
    with pytest.raises(Exception, match="Unknown failure_reason foo - please file a bug!") as e:
        list(c.increase_failure_reason_metric(failure_reason="foo", labels={}))
    assert str(e.value) == "Unknown failure_reason foo - please file a bug!"


def test_invalid_config_type(query):
    """Test exception when passing the wrong config type to DNSCollector()."""
    with pytest.raises(ConfigTypeError):
        DNSCollector(42, query, {})


def test_invalid_query_type(config):
    """Test exception when passing the wrong query type to DNSCollector()."""
    with pytest.raises(QueryTypeError):
        DNSCollector(config, 42, {})


def test_invalid_labels_type(config, query):
    """Test exception when passing the wrong labels type to DNSCollector()."""
    with pytest.raises(LabelsTypeError):
        DNSCollector(config, query, 42)


def test_get_dns_response_connectionrefusederror(
    config, query, labels, mock_get_dns_response_connectionrefusederror, caplog
):
    """Test exception when get_dns_response() raises ConnectionRefusedError."""
    caplog.set_level(logging.DEBUG)
    c = DNSCollector(config, query, labels)
    list(c.collect_dns())
    assert "failure reason is 'connection_error'" in caplog.text


def test_get_dns_response_oserror(config, query, labels, mock_get_dns_response_oserror, caplog):
    """Test exception when get_dns_response() raises OSError."""
    caplog.set_level(logging.DEBUG)
    c = DNSCollector(config, query, labels)
    list(c.collect_dns())
    assert "failure reason is 'connection_error'" in caplog.text



"""Certificate related tests."""

import logging
import ssl

import pytest
import requests


@pytest.mark.parametrize("protocol", ["dot", "doh", "doh3", "doq"])
def test_cert_verify_fail(dns_exporter_example_config, protocol, caplog):
    """Test cert verify functionality when there is a certificate<>hostname mismatch."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": protocol,
            "family": "ipv4",
            "ip": "94.140.14.140",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


@pytest.mark.parametrize("protocol", ["dot", "doh", "doh3", "doq"])
def test_cert_verify_fail_custom_ca(dns_exporter_example_config, protocol, caplog):
    """Test cert verify functionality with a selfsigned cert as CA."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns-unfiltered.adguard.com",
            "query_name": "example.com",
            "protocol": protocol,
            "family": "ipv4",
            "verify_certificate_path": "tests/certificates/test.crt",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


@pytest.mark.parametrize("protocol", ["dot", "doh", "doh3", "doq"])
def test_cert_verify_false(dns_exporter_example_config, protocol, caplog):
    """Test cert verify functionality disabled allows lookups with bad certs for doh."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "94.140.14.141",  # dns-unfiltered.adguard.com
            "query_name": "example.com",
            "protocol": protocol,
            "family": "ipv4",
            "verify_certificate": False,
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


@pytest.mark.parametrize("protocol", ["dot", "doh", "doh3", "doq"])
def test_cert_verify_invalid_path(dns_exporter_example_config, protocol, caplog):
    """Test cert verify functionality when an invalid CA path is provided."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "91.239.100.100",
            "query_name": "example.com",
            "protocol": protocol,
            "family": "ipv4",
            "verify_certificate_path": "/nonexistant",
        },
        timeout=5,
    )
    assert "FailCollector returning failure reason: invalid_request_path" in caplog.text
    assert "dnsexp_dns_query_success 0.0" in r.text


@pytest.mark.parametrize(
    "protocol",
    [
        "dot",
        "doh",
        "doh3",
        "doq",
    ],
)
def test_cert_verify_custom_ca_dir(dns_exporter_example_config, protocol, caplog):
    """Test cert verify functionality with a custom CA dir."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns-unfiltered.adguard.com",
            "query_name": "example.com",
            "protocol": protocol,
            "family": "ipv4",
            "verify_certificate_path": ssl.get_default_verify_paths().capath,
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text



"""``dns_exporter.socket_cache`` contains connection reuse related code.

The socket cache is only used when connection reuse is enabled for a query.
"""
# mypy: disable-error-code="no-untyped-call"
# ruff: noqa: SLF001

from __future__ import annotations

import logging
import socket
import threading
import time
from collections import deque
from dataclasses import dataclass
from typing import TYPE_CHECKING, NamedTuple

import dns.query
import dns.quic
import httpx
from typing_extensions import Self

from dns_exporter.metrics import (
    dnsexp_socket_age_seconds,
    dnsexp_socket_idle_seconds,
    dnsexp_socket_receive_bytes_total,
    dnsexp_socket_transmit_bytes_total,
    dnsexp_socket_uses_total,
    dnsexp_sockets_total,
)

if TYPE_CHECKING:
    import ssl
    import urllib.parse
    from collections.abc import Mapping

    from dns.quic._sync import SyncQuicConnection

    from dns_exporter.config import Config

logger = logging.getLogger(f"dns_exporter.{__name__}")


@dataclass
class SocketCacheKey:
    """Dict keys used to identify sockets in the SocketCache."""

    protocol: str
    server: str
    ip: str
    verify: str
    proxy: str

    def __eq__(self, other: object) -> bool:
        """Consider two cache keys identical only if all values match."""
        if isinstance(other, SocketCacheKey):
            return all(
                [
                    self.protocol == other.protocol,
                    self.server == other.server,
                    self.ip == other.ip,
                    self.verify == other.verify,
                    self.proxy == other.proxy,
                ]
            )
        return False

    def __hash__(self) -> int:
        """Hash all values to make these suitable for use as dict keys."""
        return hash((self.protocol, self.server, self.ip, self.verify, self.proxy))

    @property
    def labels(self) -> tuple[str, str, str, str, str]:
        """Return socket values in a format suitable for use as prometheus labels."""
        return (
            self.protocol,
            self.server,
            self.ip,
            self.verify,
            self.proxy,
        )

    @property
    def kind(self) -> socket.SocketKind:
        """Return the kind of socket."""
        return socket.SOCK_STREAM if self.protocol in ["tcp", "dot", "doh"] else socket.SOCK_DGRAM


@dataclass(kw_only=True)
class BaseSocket:
    """Baseclass used by the real classes each representing a single reusable socket."""

    create_timestamp: float | None = None
    last_use_timestamp: float | None = None
    serial: int
    cachekey: SocketCacheKey
    lock: threading.Lock
    use_count: int = 0
    bytes_sent: int = 0
    bytes_received: int = 0

    def __post_init__(self) -> None:
        """Set socket creation time."""
        if not self.create_timestamp:
            self.create_timestamp = time.time()

    def register_use(self, bytes_sent: int, bytes_received: int) -> None:
        """Called whenever a socket was used. Update use_count, last_use_time, and byte counters."""
        self.use_count += 1
        self.last_use_timestamp = time.time()
        self.bytes_sent += bytes_sent
        self.bytes_received += bytes_received


@dataclass
class PlainSocket(BaseSocket):
    """Plaintext TCP or UDP, v4 or v6 reusable socket."""

    socket: socket.socket


class PlainSockLock(NamedTuple):
    """Tuple of a deque() of PlainSockets and a lock to use when adding/reordering/deleting sockets in the deque()."""

    sockets: deque[PlainSocket]
    lock: threading.Lock


@dataclass
class DoTSocket(BaseSocket):
    """DoT v4 or v6 reusable socket."""

    socket: ssl.SSLSocket


class DoTSockLock(NamedTuple):
    """Tuple of a deque() of DoTSockets and a lock to use when adding/reordering/deleting sockets in the deque()."""

    sockets: deque[DoTSocket]
    lock: threading.Lock


@dataclass
class DoHSocket(BaseSocket):
    """DoH v4 or v6 reusable socket."""

    socket: httpx.Client


class DoHSockLock(NamedTuple):
    """Tuple of a deque() of DoHSockets and a lock to use when adding/reordering/deleting sockets in the deque()."""

    sockets: deque[DoHSocket]
    lock: threading.Lock


@dataclass
class QUICSocket(BaseSocket):
    """DoQ or DoH3 v4 or v6 reusable socket."""

    socket: SyncQuicConnection


class QUICSockLock(NamedTuple):
    """Tuple of a deque() of QUICSockets and a lock to use when adding/reordering/deleting sockets in the deque()."""

    sockets: deque[QUICSocket]
    lock: threading.Lock


class Singleton:
    """Thread safe singleton class used by SocketCache."""

    _instance = None
    _lock = threading.Lock()

    def __new__(cls) -> Self:
        """Make sure we only ever have one instance of this class."""
        if cls._instance is None:
            with cls._lock:
                # Another thread could have created the instance
                # before we acquired the lock. So check that the
                # instance is still nonexistent.
                if not cls._instance:
                    cls._instance = super().__new__(cls)
        return cls._instance


class SocketCache(Singleton):
    """Singleton socket cache."""

    # protocols tcp+udp
    plain_sockets: dict[SocketCacheKey, PlainSockLock]
    # protocol dot
    dot_sockets: dict[SocketCacheKey, DoTSockLock]
    # protocol doh
    doh_sockets: dict[SocketCacheKey, DoHSockLock]
    # protocols doq+doh3
    quic_sockets: dict[SocketCacheKey, QUICSockLock]

    def __init__(self) -> None:
        """Initialise socket cache."""
        self.plain_sockets = {}
        self.dot_sockets = {}
        self.doh_sockets = {}
        self.quic_sockets = {}
        # max. socket age
        self.socket_max_age_seconds: int = 0
        # max. socket idle time
        self.socket_max_idle_seconds: int = 3600
        # housekeeping interval
        self.housekeeping_interval: int = 600
        # set this to ask the socketcache to exit the housekeeping loop
        self.housekeeping_exit_event = threading.Event()

    def get_cache_key(self, config: Config, force_protocol: str = "") -> SocketCacheKey:
        """Return a SocketCacheKey for identifying a socket."""
        protocol = force_protocol if force_protocol else config.protocol
        if TYPE_CHECKING:  # pragma: no cover
            assert isinstance(config.server, urllib.parse.SplitResult)
        return SocketCacheKey(
            protocol=protocol,
            server=config.server.geturl().replace("udptcp://", f"{protocol}://", 1),
            ip=str(config.ip),
            verify="none" if config.verify is None else str(config.verify),
            proxy=config.proxy.geturl() if config.proxy else "none",
        )

    def get_plaintext_socket(self, *, config: Config, force_protocol: str = "") -> tuple[PlainSocket, bool]:
        """Create+connect new/return existing locked+connected TCP or UDP PlainSocket object."""
        protocol = force_protocol if force_protocol else config.protocol
        cachekey = self.get_cache_key(config=config, force_protocol=protocol)
        # check for existing socket
        if cachekey not in self.plain_sockets:
            self.plain_sockets[cachekey] = PlainSockLock(sockets=deque(), lock=threading.Lock())
        sock_lock: PlainSockLock = self.plain_sockets[cachekey]

        # loop over existing sockets
        for sock in list(sock_lock.sockets):
            # looking for an unlocked socket
            if sock.lock.acquire(blocking=False):
                # socket lock acquired, lock the deque and move this socket to the end
                with sock_lock.lock:
                    sock_lock.sockets.remove(sock)
                    sock_lock.sockets.append(sock)
                return sock, True

        # no available socket found, lock the deque and add a new socket
        with sock_lock.lock:
            # sockets are numbered from 0 and up, find an unused serial
            i = 0
            while i in [s.serial for s in sock_lock.sockets]:
                i += 1
            # create and lock socket
            sock = PlainSocket(
                socket=dns.query.make_socket(af=config.socket_family, type=cachekey.kind),
                serial=i,
                cachekey=cachekey,
                lock=threading.Lock(),
            )
            sock.lock.acquire()
            # append new socket to the end of the deque
            sock_lock.sockets.append(sock)
        if protocol == "tcp":
            self.connect_tcp_socket(config=config, sock=sock.socket)
        return sock, False

    def get_dot_socket(self, *, config: Config, verify: str | bool) -> tuple[DoTSocket, bool]:
        """Create+connect new or return existing DoTSocket."""
        cachekey = self.get_cache_key(config=config)
        if cachekey not in self.dot_sockets:
            # this is the first time this cachekey is seen, initialise deque and lock
            self.dot_sockets[cachekey] = DoTSockLock(sockets=deque(), lock=threading.Lock())
        sock_lock: DoTSockLock = self.dot_sockets[cachekey]

        # loop over existing sockets
        for sock in list(sock_lock.sockets):
            # looking for an unlocked socket
            if sock.lock.acquire(blocking=False):
                # socket lock acquired, lock the deque and move this socket to the end
                with sock_lock.lock:
                    sock_lock.sockets.remove(sock)
                    sock_lock.sockets.append(sock)
                return sock, True

        # no available socket found, lock the deque and add a new socket
        with sock_lock.lock:
            # sockets are numbered from 0 and up, find an unused serial
            i = 0
            while i in [s.serial for s in sock_lock.sockets]:
                i += 1
            # create and lock socket
            context = dns.query.make_ssl_context(verify=verify, check_hostname=bool(verify), alpns=["dot"])
            sock = DoTSocket(
                socket=dns.query.make_ssl_socket(
                    af=config.socket_family,
                    type=socket.SOCK_STREAM,
                    ssl_context=context,
                    server_hostname=config.server.hostname if config.server and verify else None,
                ),
                serial=i,
                cachekey=cachekey,
                lock=threading.Lock(),
            )
            sock.lock.acquire()
            # append new socket to the end of the deque
            sock_lock.sockets.append(sock)
        # connect and TLS handshake socket and return
        self.connect_tcp_socket(config=config, sock=sock.socket)
        dns.query._tls_handshake(s=sock.socket, expiration=time.time() + config.timeout)
        return sock, False

    def get_doh_socket(self, *, config: Config, verify: str | ssl.SSLContext | bool) -> tuple[DoHSocket, bool]:
        """Create+connect new or return existing DoHSocket."""
        cachekey = self.get_cache_key(config=config)
        if cachekey not in self.doh_sockets:
            # this is the first time this cachekey is seen, initialise deque and lock
            self.doh_sockets[cachekey] = DoHSockLock(sockets=deque(), lock=threading.Lock())
        sock_lock = self.doh_sockets[cachekey]

        # loop over existing sockets
        for sock in list(sock_lock.sockets):
            # looking for an unlocked socket
            if sock.lock.acquire(blocking=False):
                # socket lock acquired, lock the deque and move this socket to the end
                with sock_lock.lock:
                    sock_lock.sockets.remove(sock)
                    sock_lock.sockets.append(sock)
                return sock, True

        # no available socket found, lock the deque and add a new socket
        with sock_lock.lock:
            # sockets are numbered from 0 and up, find an unused serial
            i = 0
            while i in [s.serial for s in sock_lock.sockets]:
                i += 1
            # create and lock socket
            transport = dns.query._HTTPTransport(
                http1=False,
                http2=True,
                verify=verify,
                bootstrap_address=str(config.ip),
                family=config.family,
            )
            sock = DoHSocket(
                socket=httpx.Client(http1=False, http2=True, verify=verify, transport=transport),
                serial=i,
                cachekey=cachekey,
                lock=threading.Lock(),
            )
            sock.lock.acquire()
            # append new socket to the end of the deque
            sock_lock.sockets.append(sock)
        return sock, False

    def get_quic_socket(self, *, config: Config, verify: str | bool) -> tuple[QUICSocket, bool]:
        """Create+connect new or return existing QUIC connection for doq or doh3."""
        cachekey = self.get_cache_key(config=config)
        if cachekey not in self.quic_sockets:
            # this is the first time this cachekey is seen, initialise deque and lock
            self.quic_sockets[cachekey] = QUICSockLock(sockets=deque(), lock=threading.Lock())
        logger.debug(f"Inside get_quic_socket with cachekey {cachekey}")
        sock_lock: QUICSockLock = self.quic_sockets[cachekey]

        # loop over existing sockets
        for sock in list(sock_lock.sockets):
            if sock.socket._done:
                # this socket has been closed by remote end, delete and continue
                logger.debug(f"Deleting stale QUIC socket {sock}")
                self.delete_socket(sock=sock)
                continue
            # looking for an unlocked socket
            if sock.lock.acquire(blocking=False):
                # socket lock acquired, lock the deque and move this socket to the end
                with sock_lock.lock:
                    sock_lock.sockets.remove(sock)
                    sock_lock.sockets.append(sock)
                return sock, True

        # no available socket found, lock the deque and add a new socket
        with sock_lock.lock:
            # sockets are numbered from 0 and up, find an unused serial
            i = 0
            while i in [s.serial for s in sock_lock.sockets]:
                i += 1
            # create and lock a new socket
            manager = dns.quic.SyncQuicManager(
                verify_mode=verify,
                server_name=config.server.hostname if config.server else None,
                h3=config.protocol == "doh3",
            )
            sock = QUICSocket(socket=manager.connect(*config.dest), serial=i, cachekey=cachekey, lock=threading.Lock())
            sock.lock.acquire()
            # append new socket to the end of the deque
            sock_lock.sockets.append(sock)
        return sock, False

    def connect_tcp_socket(self, config: Config, sock: socket.socket | ssl.SSLSocket) -> None:
        """Connect and handshake TCP sockets. Used for protocols tcp and dot."""
        cachekey = self.get_cache_key(config=config)
        logger.debug(f"Connecting {cachekey} socket to {config.dest} ...")
        dns.query._connect(
            s=sock,
            address=config.dest,
            expiration=time.time() + config.timeout,
        )
        # set TCP_NODELAY on TCP sockets to make it go brrr
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    def update_metrics(self) -> None:
        """Update metrics for the SocketCache."""
        for sockettype in ["plain_sockets", "dot_sockets", "quic_sockets", "doh_sockets"]:
            sockets = getattr(self, sockettype)
            logger.debug(f"Updating SocketCache metrics for {len(sockets)} {sockettype}")
            for key, sock_lock in sockets.items():
                # get the number of identical sockets
                dnsexp_sockets_total.labels(*key.labels).set(len(sock_lock.sockets))
                # loop over each and get metrics for the specific socket
                for sock in list(sock_lock.sockets):
                    # include sock.serial in labels for the rest of the metrics
                    labels = [*key.labels, str(sock.serial)]
                    dnsexp_socket_age_seconds.labels(*labels).set(time.time() - sock.create_timestamp)
                    if sock.last_use_timestamp:
                        dnsexp_socket_idle_seconds.labels(*labels).set(time.time() - sock.last_use_timestamp)
                    dnsexp_socket_transmit_bytes_total.labels(*labels).set(sock.bytes_sent)
                    dnsexp_socket_receive_bytes_total.labels(*labels).set(sock.bytes_received)
                    dnsexp_socket_uses_total.labels(*labels).set(sock.use_count)

    def delete_socket(
        self, *, sock: PlainSocket | DoTSocket | DoHSocket | QUICSocket, delete_metrics: bool = True
    ) -> None:
        """Delete a socket from the cache."""
        if isinstance(sock, PlainSocket):
            try:
                # be nice and close sockets properly
                sock.socket.shutdown(socket.SHUT_RDWR)
                sock.socket.close()
            except OSError:
                # Ignore OSErrors here (socket might be closed already) and delete the socket anyway
                pass
            # remove socket from deque
            with self.plain_sockets[sock.cachekey].lock:
                self.plain_sockets[sock.cachekey].sockets.remove(sock)

        if isinstance(sock, DoTSocket):
            try:
                # be nice and close sockets right
                sock.socket.shutdown(how=socket.SHUT_RDWR)
                sock.socket.close()
            except OSError:
                # Ignore OSErrors here (socket might be closed already) and delete the socket anyway
                pass
            # remove socket from deque
            with self.dot_sockets[sock.cachekey].lock:
                self.dot_sockets[sock.cachekey].sockets.remove(sock)

        if isinstance(sock, DoHSocket):
            sock.socket.close()
            # remove socket from deque
            with self.doh_sockets[sock.cachekey].lock:
                self.doh_sockets[sock.cachekey].sockets.remove(sock)

        if isinstance(sock, QUICSocket):
            sock.socket.close()
            # remove socket from deque
            with self.quic_sockets[sock.cachekey].lock:
                self.quic_sockets[sock.cachekey].sockets.remove(sock)

        if delete_metrics:
            self.delete_metric(sock=sock)

    def delete_all_sockets(self, *, delete_metrics: bool = True) -> None:
        """Delete all sockets from socket cache."""
        for socklock in [
            *self.plain_sockets.values(),
            *self.dot_sockets.values(),
            *self.doh_sockets.values(),
            *self.quic_sockets.values(),
        ]:
            for sock in list(socklock.sockets):  # type: ignore[attr-defined]
                self.delete_socket(sock=sock, delete_metrics=delete_metrics)

    def delete_metric(self, sock: PlainSocket | DoTSocket | DoHSocket | QUICSocket) -> None:
        """Delete metrics for a socketcache entry."""
        dnsexp_sockets_total.remove(*sock.cachekey.labels)
        labels = [*sock.cachekey.labels, sock.serial]
        dnsexp_socket_age_seconds.remove(*labels)
        dnsexp_socket_idle_seconds.remove(*labels)
        dnsexp_socket_transmit_bytes_total.remove(*labels)
        dnsexp_socket_receive_bytes_total.remove(*labels)
        dnsexp_socket_uses_total.remove(*labels)

    def housekeeping(self) -> None:
        """Housekeeping method to clean old or idle sockets."""
        # loop forever
        while True:
            # start by waiting the configured interval
            if self.housekeeping_exit_event.wait(timeout=self.housekeeping_interval):
                # exit was requested, break out of the loop
                break
            # do the housekeeping
            logger.debug("SocketCache cleanup task running...")
            self.cleanup_sockets(sockets=self.plain_sockets)
            self.cleanup_sockets(sockets=self.dot_sockets)
            self.cleanup_sockets(sockets=self.doh_sockets)
            self.cleanup_sockets(sockets=self.quic_sockets)
            # log a message and retart the loop
            logger.debug(f"SocketCache cleanup task done, will run again in {self.housekeeping_interval} seconds.")

    def cleanup_sockets(
        self, sockets: Mapping[SocketCacheKey, PlainSockLock | DoTSockLock | DoHSockLock | QUICSockLock]
    ) -> None:
        """Do housekeeping for a dict of sockets."""
        now = time.time()
        keys = list(sockets.keys())
        for cachekey in keys:
            for dnssocket in list(sockets[cachekey].sockets):
                if TYPE_CHECKING:
                    assert dnssocket.create_timestamp is not None
                    assert dnssocket.last_use_timestamp is not None
                    assert isinstance(dnssocket, PlainSocket | DoTSocket | DoHSocket | QUICSocket)
                # check socket age?
                if self.socket_max_age_seconds:
                    socket_age = int(now - dnssocket.create_timestamp)
                    if socket_age > self.socket_max_age_seconds:
                        logger.debug(
                            f"Deleting socket {cachekey} index {dnssocket.serial} due to age "
                            f"{socket_age} seconds > max. age {self.socket_max_age_seconds} seconds"
                        )
                        self.delete_socket(sock=dnssocket)
                        continue

                # check socket idle time?
                if self.socket_max_idle_seconds and dnssocket.last_use_timestamp:
                    socket_idle = now - dnssocket.last_use_timestamp
                    if socket_idle > self.socket_max_idle_seconds:
                        logger.debug(
                            f"Deleting socket {cachekey} index {dnssocket.serial} with "
                            f"idle time {socket_idle} seconds > max. idle "
                            f"{self.socket_max_idle_seconds} seconds"
                        )
                        self.delete_socket(sock=dnssocket)


def cleanup_socket_cache(socket_cache: SocketCache, housekeeping_thread: threading.Thread | None) -> None:
    """Close and delete all sockets and stop the housekeeping thread. Used on exit."""
    logger.debug("SocketCache cleanup running")
    if housekeeping_thread:
        logger.debug("Asking housekeeping thread to exit...")
        socket_cache.housekeeping_exit_event.set()
        logger.debug("Waiting for SocketCache housekeeping thread to exit...")
        housekeeping_thread.join()
        logger.debug("The SocketCache housekeeping thread exited cleanly.")
    if (
        len(socket_cache.plain_sockets)
        + len(socket_cache.dot_sockets)
        + len(socket_cache.doh_sockets)
        + len(socket_cache.quic_sockets)
    ):
        logger.debug(
            "Closing and deleting "
            f"{len(socket_cache.plain_sockets)} plain sockets, "
            f"{len(socket_cache.dot_sockets)} DoT sockets, "
            f"{len(socket_cache.doh_sockets)} DoH sockets, and "
            f"{len(socket_cache.quic_sockets)} QUIC sockets in the SocketCache..."
        )
        # do not waste time deleting metrics before exiting, just close sockets
        socket_cache.delete_all_sockets(delete_metrics=False)
    logger.debug("SocketCache cleanup done.")



"""``dns_exporter.collector`` contains the DNSCollector class used by the DNSExporter during scrapes."""

from __future__ import annotations

import contextlib
import logging
import re
import socket
import ssl
import time
import traceback
import urllib
from dataclasses import dataclass
from typing import TYPE_CHECKING

import dns.edns
import dns.exception
import dns.flags
import dns.inet
import dns.opcode
import dns.query
import dns.quic
import dns.rcode
import dns.rdatatype
import dns.resolver
import httpx
import socks  # type: ignore[import-untyped]
from dns.message import Message, QueryMessage
from prometheus_client.core import CounterMetricFamily, GaugeMetricFamily
from prometheus_client.registry import Collector

from dns_exporter.config import Config
from dns_exporter.exceptions import (
    ConfigTypeError,
    LabelsTypeError,
    ProtocolSpecificError,
    QueryTypeError,
    UnknownFailureReasonError,
    ValidationError,
)
from dns_exporter.metrics import (
    FAILURE_REASONS,
    TTL_LABELS,
    dnsexp_dns_queries_total,
    dnsexp_dns_responsetime_seconds,
    dnsexp_scrape_failures_total,
    get_dns_qtime_metric,
    get_dns_success_metric,
    get_dns_ttl_metric,
)
from dns_exporter.socket_cache import DoHSocket, DoTSocket, PlainSocket, QUICSocket, SocketCache
from dns_exporter.version import __version__

if TYPE_CHECKING:  # pragma: no cover
    import urllib.parse
    from collections.abc import Iterator
    from ipaddress import IPv4Address, IPv6Address

    from dns.message import Message

    from dns_exporter.config import RRValidator

logger = logging.getLogger(f"dns_exporter.{__name__}")

socket_cache = SocketCache()


@dataclass
class DNSResponse:
    """Class to hold the result of a lookup."""

    message: Message | None
    transport: str  # TCP, UDP, or QUIC
    socket_reused: bool

    def __bool__(self) -> bool:
        """Be falsy when there is no message."""
        return bool(self.message)


class DNSCollector(Collector):
    """Custom collector class which does DNS lookups and returns metrics."""

    # set the version on the class
    __version__: str = __version__

    # used to store a reference to the socket during queries
    sock: PlainSocket | DoTSocket | DoHSocket | QUICSocket

    def __init__(
        self,
        config: Config,
        query: QueryMessage,
        labels: dict[str, str],
    ) -> None:
        """Save config and q object as class attributes for use later."""
        logger.debug("Initialising DNSCollector...")
        # make sure config is valid
        if isinstance(config, Config):
            self.config = config
        else:
            raise ConfigTypeError

        # make sure query is valid
        if isinstance(query, QueryMessage):
            self.query = query
        else:
            raise QueryTypeError

        # make sure labels is a list of strings
        if isinstance(labels, dict):
            self.labels = {k: str(v) for k, v in labels.items()}
        else:
            raise LabelsTypeError

        # set proxy?
        if self.config.proxy:
            socks.set_default_proxy(
                proxy_type=getattr(socks, self.config.proxy.scheme.upper()),
                addr=self.config.proxy.hostname,
                port=self.config.proxy.port,
            )
            if self.config.protocol in ["doh3", "doq"]:
                dns.quic._sync.socket_factory = socks.socksocket  # noqa: SLF001
            else:
                dns.query.socket_factory = socks.socksocket
            logger.debug(f"Using proxy {self.config.proxy.geturl()}")
        else:
            # no proxy, make sure sockets are reset to socket.socket
            dns.query.socket_factory = socket.socket
            dns.quic._sync.socket_factory = socket.socket  # noqa: SLF001
            logger.debug("Not using a proxy for this request")

    def describe(self) -> Iterator[CounterMetricFamily | GaugeMetricFamily]:
        """Describe the metrics that are to be returned by this collector."""
        yield get_dns_qtime_metric()
        yield get_dns_success_metric()
        yield get_dns_ttl_metric()
        yield from self.collect_up()

    def collect(
        self,
    ) -> Iterator[CounterMetricFamily | GaugeMetricFamily]:
        """Do DNS lookup and yield metrics."""
        yield from self.collect_dns()
        yield from self.collect_up()
        logger.debug("Done, returning HTTP response")

    def collect_up(self) -> Iterator[GaugeMetricFamily]:
        """Yield the up metric."""
        yield GaugeMetricFamily(
            "up",
            "The value of this Gauge is always 1 when the dns_exporter is up",
            value=1,
        )

    def collect_dns(self) -> Iterator[CounterMetricFamily | GaugeMetricFamily]:
        """Collect and yield DNS metrics."""
        # pleasing mypy
        if TYPE_CHECKING:  # pragma: no cover
            assert isinstance(self.config.ip, IPv4Address | IPv6Address)
            assert isinstance(self.config.server, urllib.parse.SplitResult)
            assert isinstance(self.config.server.port, int)

        r = None
        # mark the start time and do the request
        start = time.time()
        try:
            r = self.get_dns_response()
            logger.debug(
                f"Protocol {self.config.protocol} got a DNS query response over {r.transport}",
            )
        except dns.exception.Timeout:
            # configured timeout was reached before a response arrived
            reason = "timeout"
            self.increase_failure_reason_metric(failure_reason=reason, labels=self.labels)
        except ConnectionRefusedError:
            # server actively refused the connection
            reason = "connection_error"
            self.increase_failure_reason_metric(
                failure_reason=reason,
                labels=self.labels,
            )
        except OSError as e:
            # raised by multiple protocols on ICMP unreach
            logger.debug(f"Protocol {self.config.protocol} got OSError '{e}', exception follows", exc_info=True)
            reason = "connection_error"
            self.increase_failure_reason_metric(
                failure_reason=reason,
                labels=self.labels,
            )
        except ProtocolSpecificError as e:
            # a protocol specific exception was raised, log and re-raise
            logger.debug(
                f"Protocol {self.config.protocol} raised exception, returning failure reason {e}",
                exc_info=True,
            )
            reason = str(e)
            self.increase_failure_reason_metric(failure_reason=reason, labels=self.labels)
        except Exception:  # noqa: BLE001
            logger.warning(
                f"""Caught an unknown exception while looking up qname {self.config.query_name} using server
                {self.config.server.geturl()} and proxy {self.config.proxy.geturl() if self.config.proxy else "none"}
                - exception details follow, returning other_failure""",
                exc_info=True,
            )
            reason = "other_failure"
            self.increase_failure_reason_metric(failure_reason=reason, labels=self.labels)

        # clock it
        qtime = time.time() - start

        # did we get a response?
        if not r or r.message is None:
            logger.debug(
                f"No DNS response received from server {self.config.server.geturl()} - failure reason is '{reason}'..."
            )
            yield from (
                get_dns_qtime_metric(),
                get_dns_ttl_metric(),
                get_dns_success_metric(value=0),
            )
            return None

        # parse response (if any) and yield metrics
        yield from self.handle_response(
            response=r.message, transport=r.transport, connection_reused=r.socket_reused, qtime=qtime
        )

    def handle_response(
        self, *, response: Message, transport: str, connection_reused: bool, qtime: float
    ) -> Iterator[CounterMetricFamily | GaugeMetricFamily]:
        """Do response processing and yield metrics."""
        # convert response flags to sorted text
        flags = dns.flags.to_text(response.flags).split(" ")
        flags.sort()

        # update labels with data from the response
        self.labels.update(
            {
                "transport": transport,
                "opcode": dns.opcode.to_text(response.opcode()),
                "rcode": dns.rcode.to_text(response.rcode()),
                "flags": " ".join(flags),
                "answer": str(sum([len(rrset) for rrset in response.answer])),
                "authority": str(len(response.authority)),
                "additional": str(len(response.additional)),
                "nsid": "no_nsid",
            },
        )

        # does the answer have nsid?
        self.handle_response_options(response=response)

        # include connection label?
        if "connection" in self.labels:  # pragma: no cover
            self.labels["connection"] = "reused" if connection_reused else "new"

        # labels complete, yield timing metric
        qtime_metric = get_dns_qtime_metric()
        qtime_metric.add_metric(labels=list(self.labels.values()), value=qtime)
        yield qtime_metric

        # update internal exporter metric
        dnsexp_dns_responsetime_seconds.labels(**self.labels).observe(qtime)

        yield from self.yield_ttl_metrics(response=response)

        # validate response and yield remaining metrics
        logger.debug("Validating response and yielding remaining metrics")
        try:
            self.validate_response(response=response)
            self.increase_failure_reason_metric(failure_reason="", labels=self.labels)
            yield get_dns_success_metric(1)
        except ValidationError as E:
            logger.debug(f"Validation failed: {E.args[1]}")
            self.increase_failure_reason_metric(failure_reason=E.args[1], labels=self.labels)
            yield get_dns_success_metric(0)

    def handle_response_options(self, response: Message) -> None:
        """Handle response edns."""
        for opt in response.options:
            if opt.otype == dns.edns.NSID:
                nsid = opt.to_text()
                if nsid.startswith("NSID"):
                    nsid = nsid[5:]
                # do we have an NSID string? then overwrite the default 'no_nsid' string
                if nsid:
                    self.labels.update({"nsid": nsid})
                break

    def yield_ttl_metrics(self, response: Message) -> Iterator[GaugeMetricFamily]:
        """Register TTL of response RRs and yield ttl metric."""
        ttl = get_dns_ttl_metric()
        if self.config.collect_ttl:
            for section in ["answer", "authority", "additional"]:
                logger.debug(f"processing section {section}")
                rrsets = getattr(response, section)
                for rrset in rrsets:
                    logger.debug(f"processing rrset {rrset}...")
                    for rr in rrset:
                        logger.debug(f"processing rr {rr}")
                        self.labels.update(
                            {
                                "rr_section": section,
                                "rr_name": str(rrset.name),
                                "rr_type": dns.rdatatype.to_text(rr.rdtype),
                                "rr_value": rr.to_text()[: self.config.collect_ttl_rr_value_length],
                            },
                        )
                        ttl.add_metric(list(self.labels.values()), rrset.ttl)
        # yield all the ttl metrics
        logger.debug("yielding ttl metrics")
        yield ttl

    def get_dns_response(  # noqa: PLR0911 PLR0912 C901
        self,
        *,
        retry: bool = False,
    ) -> DNSResponse:
        """Perform a DNS query with the configured server and protocol."""
        if not retry:
            # increase query counter
            dnsexp_dns_queries_total.inc()

        # get proxy string for logging
        proxy = self.config.proxy.geturl() if self.config.proxy else "is not active"

        if TYPE_CHECKING:  # pragma: no cover
            # please mypy
            assert self.config.server is not None
            assert isinstance(self.config.server.port, int)

        # prepare variables
        ip = str(self.config.ip)
        port = int(self.config.server.port)

        logger.debug(
            f"Doing DNS query {self.query.question} with server {self.config.server.geturl()} "
            f"(using IP {self.config.ip}) and proxy {proxy}"
        )

        try:
            if self.config.protocol == "udp":
                # plain UDP lookup, nothing fancy here
                return self.get_dns_response_udp(
                    ip=ip,
                    port=port,
                )

            if self.config.protocol == "tcp":
                # plain TCP lookup, nothing fancy here
                return self.get_dns_response_tcp(
                    ip=ip,
                    port=port,
                )

            if self.config.protocol == "udptcp":
                # plain UDP lookup with fallback to TCP lookup
                return self.get_dns_response_udptcp(
                    ip=ip,
                    port=port,
                )

            if self.config.protocol == "dot":
                return self.get_dns_response_dot(
                    ip=ip,
                    port=port,
                    server=self.config.server,
                    verify=self.config.get_tls_verify(),
                )

            if self.config.protocol == "doh":
                return self.get_dns_response_doh(
                    ip=ip,
                    port=port,
                    server=self.config.server,
                    verify=self.config.get_tls_context(),
                )

            if self.config.protocol == "doh3":
                return self.get_dns_response_doh3(
                    ip=ip,
                    port=port,
                    server=self.config.server,
                    verify=self.config.get_tls_verify(),
                )

            if self.config.protocol == "doq":
                return self.get_dns_response_doq(
                    ip=ip,
                    port=port,
                    server=self.config.server,
                    verify=self.config.get_tls_verify(),
                )

        except (EOFError, OSError) as e:
            # EOFError can happen when reusing an old socket that has been closed from
            # the remote end, but might also in rare cases happen for new sockets.
            # OSError happens under circumstances I don't fully understand. Retry those too.
            ex = "".join(traceback.format_exception_only(e)).strip()
            if not self.config.connection_reuse:
                logger.debug(
                    f"Protocol {self.config.protocol} raised {ex}, returning socket_error",
                )
                raise ProtocolSpecificError("socket_error") from e

            # if this is the second attempt then bail out now
            if retry:
                logger.debug(
                    f"Protocol {self.config.protocol} raised {ex} after retry with new socket, returning socket_error",
                )
                raise ProtocolSpecificError("socket_error") from e

            logger.debug(
                f"Protocol {self.config.protocol} raised {ex} with existing socket, retrying with new socket...",
            )
            # first attempt failed, delete existing socket
            if hasattr(self, "sock"):
                socket_cache.delete_socket(sock=self.sock)
            # try again
            return self.get_dns_response(retry=True)

        # unknown protocol, we will never get here, but mypy wants a return statement
        return DNSResponse(message=None, transport="", socket_reused=False)  # pragma: no cover

    def get_dns_response_udp(self, *, ip: str, port: int) -> DNSResponse:
        """Perform a DNS query with the udp protocol."""
        # get reusable socket?
        if self.config.connection_reuse:
            self.sock, reused = socket_cache.get_plaintext_socket(config=self.config, force_protocol="udp")
        r = None
        try:
            r = dns.query.udp(
                q=self.query,
                where=ip,
                port=port,
                timeout=self.config.timeout,
                one_rr_per_rrset=True,
                raise_on_truncation=True,
                sock=self.sock.socket if self.config.connection_reuse else None,
            )
        except dns.message.Truncated as e:
            logger.debug("Protocol udp response truncated, returning response_trucated error")
            raise ProtocolSpecificError("response_truncated") from e
        finally:
            if self.config.connection_reuse:
                # update the socket stats
                self.sock.register_use(
                    bytes_sent=len(self.query.to_wire()), bytes_received=len(r.to_wire()) if r else 0
                )
                self.sock.lock.release()
        return DNSResponse(message=r, transport="UDP", socket_reused=reused if self.config.connection_reuse else False)

    def get_dns_response_tcp(self, *, ip: str, port: int) -> DNSResponse:
        """Perform a DNS query with the tcp protocol."""
        # get reusable socket?
        if self.config.connection_reuse:
            self.sock, reused = socket_cache.get_plaintext_socket(config=self.config, force_protocol="tcp")
        r = None
        # do the query
        try:
            r = dns.query.tcp(
                q=self.query,
                where=ip,
                port=port,
                timeout=self.config.timeout,
                one_rr_per_rrset=True,
                sock=self.sock.socket if self.config.connection_reuse else None,
            )
        finally:
            if self.config.connection_reuse:
                self.sock.register_use(
                    bytes_sent=len(self.query.to_wire()), bytes_received=len(r.to_wire()) if r else 0
                )
                self.sock.lock.release()
        return DNSResponse(message=r, transport="TCP", socket_reused=reused if self.config.connection_reuse else False)

    def get_dns_response_udptcp(self, *, ip: str, port: int) -> DNSResponse:
        """Perform a DNS query with the udptcp protocol (with fallback to TCP)."""
        # do the query over UDP first
        try:
            return self.get_dns_response_udp(ip=ip, port=port)
        except ProtocolSpecificError:
            # fallback to TCP
            logger.debug("Protocol udptcp response truncated during UDP lookup, trying TCP instead")
            return self.get_dns_response_tcp(ip=ip, port=port)

    def get_dns_response_dot(
        self,
        *,
        ip: str,
        port: int,
        server: urllib.parse.SplitResult,
        verify: str | bool,
    ) -> DNSResponse:
        """Perform a DNS query with the dot protocol and catch protocol specific exceptions."""
        if self.config.connection_reuse:
            self.sock, reused = socket_cache.get_dot_socket(config=self.config, verify=verify)
        r = None
        try:
            # DoT query, use the ip for where= and set tls hostname with server_hostname=
            r = dns.query.tls(
                q=self.query,
                where=ip,
                port=port,
                server_hostname=server.hostname if verify else None,
                timeout=self.config.timeout,
                verify=verify,
                one_rr_per_rrset=True,
                sock=self.sock.socket if self.config.connection_reuse else None,  # type: ignore[arg-type]
            )
        except ssl.SSLCertVerificationError as e:
            # raised by dot on certificate verification error
            logger.debug(
                "Protocol dot raised ssl.SSLCertVerificationError, returning certificate_error",
            )
            raise ProtocolSpecificError("certificate_error") from e
        finally:
            if self.config.connection_reuse:
                self.sock.register_use(
                    bytes_sent=len(self.query.to_wire()), bytes_received=len(r.to_wire()) if r else 0
                )
                self.sock.lock.release()
        return DNSResponse(message=r, transport="TCP", socket_reused=reused if self.config.connection_reuse else False)

    def get_dns_response_doh(
        self,
        *,
        ip: str,
        port: int,
        http_version: dns.query.HTTPVersion = dns.query.HTTPVersion.HTTP_2,
        server: urllib.parse.SplitResult,
        verify: str | ssl.SSLContext | bool,
    ) -> DNSResponse:
        """Perform a DNS query with the doh protocol (tcp+http1/2), catch protocol specific exceptions."""
        if self.config.connection_reuse:
            self.sock, reused = socket_cache.get_doh_socket(config=self.config, verify=verify)
        r = None
        try:
            # DoH query, use the url for where= and use bootstrap_address= for the ip
            url = f"https://{server.hostname}{server.path}"
            r = dns.query.https(
                q=self.query,
                where=url,
                bootstrap_address=ip,
                port=port,
                timeout=self.config.timeout,
                verify=verify,
                one_rr_per_rrset=True,
                # https://github.com/tykling/dns_exporter/issues/201
                http_version=http_version,
                session=self.sock.socket if self.config.connection_reuse else None,
            )
        except httpx.ConnectError as e:
            # raised by doh on both certificate errors and other connection issues
            reason = "certificate_error" if "CERTIFICATE_VERIFY_FAILED" in str(e) else "connection_error"
            logger.debug(f"Protocol doh raised exception, returning {reason}")
            raise ProtocolSpecificError(reason) from e
        except (httpx.ConnectTimeout, httpx.ReadTimeout) as e:
            # raised by doh on timeouts
            reason = "timeout"
            logger.debug(f"Protocol doh raised exception, returning {reason}")
            raise ProtocolSpecificError(reason) from e
        except (httpx.WriteError, httpx.RemoteProtocolError) as e:
            # raised by doh when the connection is closed by the remote server
            reason = "connection_error"
            logger.debug(f"Protocol doh raised exception, returning {reason}")
            raise ProtocolSpecificError(reason) from e
        except ValueError as e:
            # raised by doh when the server response with a non-2XX HTTP status code
            logger.debug(
                "Protocol doh raised ValueErrror due to non-2XX status_code - returning invalid_response_statuscode"
            )
            raise ProtocolSpecificError("invalid_response_statuscode") from e
        finally:
            if self.config.connection_reuse:
                self.sock.register_use(
                    bytes_sent=len(self.query.to_wire()), bytes_received=len(r.to_wire()) if r else 0
                )
                self.sock.lock.release()
        return DNSResponse(
            message=r,
            transport="TCP",
            socket_reused=reused if self.config.connection_reuse else False,
        )

    def get_dns_response_doh3(
        self,
        *,
        ip: str,
        port: int,
        server: urllib.parse.SplitResult,
        verify: str | bool,
    ) -> DNSResponse:
        """Perform a DNS query with the doh3 protocol."""
        if self.config.connection_reuse:
            self.sock, reused = socket_cache.get_quic_socket(config=self.config, verify=verify)
        # DoH3 query, use the url for where= and use bootstrap_address= for the ip
        url = f"https://{server.hostname}{server.path}"
        r = None
        try:
            r = dns.query.https(
                q=self.query,
                where=url,
                bootstrap_address=ip,
                port=port,
                timeout=self.config.timeout,
                verify=verify,
                one_rr_per_rrset=True,
                http_version=dns.query.HTTPVersion.HTTP_3,
                session=self.sock.socket if self.config.connection_reuse else None,
            )
        finally:
            if self.config.connection_reuse:
                # update the socket stats
                self.sock.register_use(
                    bytes_sent=len(self.query.to_wire()), bytes_received=len(r.to_wire()) if r else 0
                )
                self.sock.lock.release()
        return DNSResponse(
            message=r,
            transport="QUIC",
            socket_reused=reused if self.config.connection_reuse else False,
        )

    def get_dns_response_doq(
        self,
        *,
        ip: str,
        port: int,
        server: urllib.parse.SplitResult,
        verify: str | bool,
    ) -> DNSResponse:
        """Perform a DNS query with the doq protocol and catch protocol specific exceptions."""
        # DoQ query, use the IP for where= and use server_hostname for the hostname
        if self.config.connection_reuse:
            self.sock, reused = socket_cache.get_quic_socket(config=self.config, verify=verify)
        r = None
        try:
            r = dns.query.quic(
                q=self.query,
                where=ip,
                port=port,
                server_hostname=server.hostname if verify else None,
                timeout=self.config.timeout,
                verify=verify,
                one_rr_per_rrset=True,
                connection=self.sock.socket if self.config.connection_reuse else None,
            )
        finally:
            if self.config.connection_reuse:
                # update the socket stats
                self.sock.register_use(
                    bytes_sent=len(self.query.to_wire()), bytes_received=len(r.to_wire()) if r else 0
                )
                self.sock.lock.release()
        return DNSResponse(message=r, transport="QUIC", socket_reused=reused if self.config.connection_reuse else False)

    def validate_response_rcode(self, response: Message) -> None:
        """Validate response RCODE."""
        # get the rcode from the respose and validate it
        rcode = dns.rcode.to_text(response.rcode())
        if rcode not in self.config.valid_rcodes:
            logger.info(f"Raising ValidationError: RCODE {rcode} not among valid_rcodes {self.config.valid_rcodes}")
            raise ValidationError(
                "rcode_validator",
                "invalid_response_rcode",
            )

    def validate_response_flags(self, response: Message) -> None:  # noqa: PLR0912 C901
        """Validate response flags."""
        # create a list of flags as text like ["QR", "AD"]
        flags = dns.flags.to_text(response.flags).split(" ")

        if self.config.validate_response_flags.fail_if_any_present:
            for flag in self.config.validate_response_flags.fail_if_any_present:
                # if any of these flags are found in the response validation fails
                if flag in flags:
                    raise ValidationError(
                        "fail_if_any_present",
                        "invalid_response_flags",
                    )

        if self.config.validate_response_flags.fail_if_all_present:
            for flag in self.config.validate_response_flags.fail_if_all_present:
                # if all these flags are found in the response then fail
                if flag not in flags:
                    break
            else:
                # all the flags are present
                raise ValidationError(
                    "fail_if_all_present",
                    "invalid_response_flags",
                )

        if self.config.validate_response_flags.fail_if_any_absent:
            for flag in self.config.validate_response_flags.fail_if_any_absent:
                # if any of these flags is missing from the response then fail
                if flag not in flags:
                    raise ValidationError(
                        "fail_if_any_absent",
                        "invalid_response_flags",
                    )

        if self.config.validate_response_flags.fail_if_all_absent:
            for flag in self.config.validate_response_flags.fail_if_all_absent:
                # if all these flags are missing from the response then fail
                if flag in flags:
                    break
            else:
                # all the flags are missing
                raise ValidationError(
                    "fail_if_all_absent",
                    "invalid_response_flags",
                )

    def check_regexes(
        self,
        validators: RRValidator,
        validator: str,
        rrs: str,
        section: str,
    ) -> None:
        """Loop over response RRs and check for regex matches."""
        logger.debug(
            f"{validator} validating {len(rrs)} rrs from {section} section "
            f"with {len(getattr(validators, validator))} regexes..."
        )

        num_matches_per_rrs = {str(rr): 0 for rr in rrs}

        for regex in getattr(validators, validator):
            p = re.compile(regex)
            for rr in rrs:
                result = p.match(str(rr))
                log_entry = f"matching '{rr}' against '{regex}': " + ("success" if result else "failed")
                logger.debug(log_entry)
                if result:
                    num_matches_per_rrs[str(rr)] = +1

        # consider request failed if any answer rr matches one of these regexes
        # -> no regex matches any rr
        if validator == "fail_if_matches_regexp" and sum(num_matches_per_rrs.values()) == 0:
            return

        # consider request failed if all answer rrs match one of these regexes
        # -> all regex dit not match at least one rr
        if validator == "fail_if_all_match_regexp" and 0 in num_matches_per_rrs.values():
            return

        # consider request failed if any answer rr does not match one of these regexes
        # -> any regex matched at least one rr
        if validator == "fail_if_not_matches_regexp" and min(num_matches_per_rrs.values()) > 0:
            return

        # consider request failed if none of the answer rrs match one of these regexes
        # -> one regex mathed at least one rr
        if validator == "fail_if_none_matches_regexp" and sum(num_matches_per_rrs.values()) > 0:
            return

        raise ValidationError(validator, f"invalid_response_{section}_rrs")

    def validate_response_rrs(self, response: Message) -> None:
        """Validate response RRs."""
        for section in ["answer", "authority", "additional"]:
            key = f"validate_{section}_rrs"
            rrs = getattr(response, section)
            if getattr(self.config, key):
                validators: RRValidator = getattr(self.config, key)
                if rrs and validators.fail_if_matches_regexp:
                    self.check_regexes(
                        validators=validators,
                        validator="fail_if_matches_regexp",
                        rrs=rrs,
                        section=section,
                    )

                if rrs and validators.fail_if_all_match_regexp:
                    self.check_regexes(
                        validators=validators,
                        validator="fail_if_all_match_regexp",
                        rrs=rrs,
                        section=section,
                    )

                if validators.fail_if_not_matches_regexp:
                    self.check_regexes(
                        validators=validators,
                        validator="fail_if_not_matches_regexp",
                        rrs=rrs,
                        section=section,
                    )

                if validators.fail_if_none_matches_regexp:
                    self.check_regexes(
                        validators=validators,
                        validator="fail_if_none_matches_regexp",
                        rrs=rrs,
                        section=section,
                    )

    def validate_response(self, response: Message) -> None:
        """Validate the DNS response using the validation config in the config."""
        # validate the response rcode?
        if self.config.valid_rcodes:
            self.validate_response_rcode(response=response)

        # validate flags?
        if self.config.validate_response_flags:
            self.validate_response_flags(response=response)

        # check response rr validation
        self.validate_response_rrs(response=response)

    @staticmethod
    def increase_failure_reason_metric(failure_reason: str, labels: dict[str, str]) -> None:
        """This method is used to maintain failure metrics.

        If an empty string is passed as failure_reason (meaning success) the failure counters will not be incremented.
        """
        # was there a failure?
        if not failure_reason:
            return

        # is it a valid failure reason?
        if failure_reason not in FAILURE_REASONS:
            # unknown failure_reason, this is a bug
            raise UnknownFailureReasonError(failure_reason)

        # delete unwelcome labels
        for key in TTL_LABELS:
            with contextlib.suppress(KeyError):
                del labels[key]

        # build a dict with reason first and the rest of the labels after
        labeldict = {"reason": failure_reason}
        labeldict.update(labels)
        logger.debug(labeldict)
        # increase the global failure counter
        dnsexp_scrape_failures_total.labels(**labeldict).inc()
        return


class FailCollector(DNSCollector):
    """Custom collector class used to handle pre-DNSCollector failures, like configuration issues."""

    def __init__(self, failure_reason: str, labels: dict[str, str]) -> None:
        """Save failure reason for use later."""
        self.reason = failure_reason
        self.labels = labels

    def collect_dns(
        self,
    ) -> Iterator[CounterMetricFamily | GaugeMetricFamily]:
        """Do not collect anything, just return the error message."""
        logger.debug(f"FailCollector returning failure reason: {self.reason}")
        yield get_dns_qtime_metric()
        yield get_dns_ttl_metric()
        yield get_dns_success_metric(value=0)
        self.increase_failure_reason_metric(failure_reason=self.reason, labels=self.labels)
